

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.charset.Charset;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Hashtable;
import java.util.List;


import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.Namespace;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

/**
 * @author albert
 *
 */
public class ElanFile
{
	private List<ElanLinkedFile> mediaFiles;
	private List<ElanLinkedFile> linkedFiles;
	
	private List<ElanTier> tiers;

	/**
	 * Constructs a new ElanFile Object.
	 */
	public ElanFile()
	{
		this.mediaFiles = new ArrayList<ElanLinkedFile>();
		this.linkedFiles = new ArrayList<ElanLinkedFile>();

		this.tiers = new ArrayList<ElanTier>();
	}
	
	/**
	 * Constructs a new ElanFile Object.
	 */
	public ElanFile(InputStream inputStream, Charset charset)
	{
		try {
			this.read(inputStream);
		} catch (JDOMException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	@SuppressWarnings("unchecked")
	public void read(InputStream inputStream) throws JDOMException, IOException
	{
		this.mediaFiles = new ArrayList<ElanLinkedFile>();
		this.linkedFiles = new ArrayList<ElanLinkedFile>();

		this.tiers = new ArrayList<ElanTier>();
		
		Document doc = new SAXBuilder().build(inputStream);
		Element root = doc.getRootElement();
		List<Element> elements;
		Hashtable<String, Long> timeslots = new Hashtable<String, Long>();
		Element buffer;
		
		// read header
		elements = root.getChild("HEADER").getChildren("MEDIA_DESCRIPTOR");
		for (Element e: elements)
		{
			ElanLinkedFile mediaFile = new ElanLinkedFile();
			mediaFile.setURL(e.getAttributeValue("MEDIA_URL"));
			mediaFile.setRelativeURL(e.getAttributeValue("RELATIVE_MEDIA_URL"));
			mediaFile.setMimeType(e.getAttributeValue("MIME_TYPE"));
			try {
				mediaFile.setTimeOrigin(Long.parseLong(e.getAttributeValue("TIME_ORIGIN")));
			} catch (Exception ex) {
//				ex.printStackTrace();
				mediaFile.setTimeOrigin(0l);
			}
			this.mediaFiles.add(mediaFile);
		}
		elements = root.getChild("HEADER").getChildren("LINKED_FILE_DESCRIPTOR");
		for (Element e: elements)
		{
			ElanLinkedFile linkedFile = new ElanLinkedFile();
			linkedFile.setURL(e.getAttributeValue("LINK_URL"));
			linkedFile.setRelativeURL(e.getAttributeValue("RELATIVE_LINK_URL"));
			linkedFile.setMimeType(e.getAttributeValue("MIME_TYPE"));
			try {
				linkedFile.setTimeOrigin(Long.parseLong(e.getAttributeValue("TIME_ORIGIN")));
			} catch (Exception ex) {
//				ex.printStackTrace();
				linkedFile.setTimeOrigin(0l);
			}
				this.linkedFiles.add(linkedFile);
			}
		
		// read time order
		elements = root.getChild("TIME_ORDER").getChildren("TIME_SLOT");
		for (Element e: elements)
			timeslots.put(e.getAttributeValue("TIME_SLOT_ID"), Long.decode(e.getAttributeValue("TIME_VALUE")));
		
		// read tiers
		elements = root.getChildren("TIER");
		for (Element e: elements)
		{
			ElanTier tier = new ElanTier();
			tier.setID(e.getAttributeValue("TIER_ID"));
			tier.setParticipant(e.getAttributeValue("PARTICIPANT"));
			tier.setAnnotator(e.getAttributeValue("ANNOTATOR"));
			
			List<Element> annotations = e.getChildren("ANNOTATION");
			for (Element a : annotations)
			{
				buffer = a.getChild("ALIGNABLE_ANNOTATION");
				ElanAnnotation annotation = new ElanAnnotation();
				annotation.setStartTime(timeslots.get(buffer.getAttributeValue("TIME_SLOT_REF1")));
				annotation.setEndTime(timeslots.get(buffer.getAttributeValue("TIME_SLOT_REF2")));
				annotation.setValue(buffer.getChildTextTrim("ANNOTATION_VALUE"));
				tier.addAnnotation(annotation);
			}
			this.tiers.add(tier);
		}
	}
	
	public void write(OutputStream outputStream) throws IOException
	{
		// build document
		Document doc = new Document();
		
		// build root element
		// <ANNOTATION_DOCUMENT xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://www.mpi.nl/tools/elan/EAFv2.6.xsd" DATE="2010-06-02T16:23:22+01:00" AUTHOR="" VERSION="2.6" FORMAT="2.6">
		Element root = new Element("ANNOTATION_DOCUMENT");
		Namespace xsins =  Namespace.getNamespace("xsi", "http://www.w3.org/2001/XMLSchema-instance");
		root.setAttribute("noNamespaceSchemaLocation", "http://www.mpi.nl/tools/elan/EAFv2.6.xsd", xsins);
		root.setAttribute("DATE", (new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssZ").format(new Date())).substring(0, 22) + ":00");
		root.setAttribute("AUTHOR", "");
		root.setAttribute("VERSION", "2.6");
		root.setAttribute("FORMAT", "2.6");
		doc.setRootElement(root);
		
		// build header
		Element header = new Element("HEADER");
		header.setAttribute("MEDIA_FILE", "");
		header.setAttribute("TIME_UNITS", "milliseconds");
		for (ElanLinkedFile file : this.mediaFiles) {
			Element elem = new Element("MEDIA_DESCRIPTOR");
			elem.setAttribute("MEDIA_URL", file.getURL());
			elem.setAttribute("RELATIVE_MEDIA_URL", file.getRelativeURL());
			elem.setAttribute("MIME_TYPE", file.getMimeType());
			elem.setAttribute("TIME_ORIGIN", "" + file.getTimeOrigin());
			header.addContent(elem);
		}
		for (ElanLinkedFile file : this.linkedFiles) {
			Element elem = new Element("LINKED_FILE_DESCRIPTOR");
			elem.setAttribute("LINK_URL", file.getURL());
			elem.setAttribute("RELATIVE_LINK_URL", file.getRelativeURL());
			elem.setAttribute("MIME_TYPE", file.getMimeType());
			elem.setAttribute("TIME_ORIGIN", "" + file.getTimeOrigin());
			header.addContent(elem);
		}
		int numAnnotations = 0;
		for (ElanTier tier : this.tiers) numAnnotations += tier.getAnnotations().size();
		header.addContent(new Element("PROPERTY").setAttribute("NAME", "lastUsedAnnotationId").setText(""+numAnnotations));
		root.addContent(header);
		
		// build time order
		Element timeOrder = new Element("TIME_ORDER");
		int numTimeSlots = 1;
		for (ElanTier tier : this.tiers) {
			List<ElanAnnotation> annotations = tier.getAnnotations();
			for (ElanAnnotation annotation : annotations) {
				Element elem = new Element("TIME_SLOT");
				elem.setAttribute("TIME_SLOT_ID", "ts" + numTimeSlots);
				elem.setAttribute("TIME_VALUE", "" + annotation.getStartTime());
				timeOrder.addContent(elem);
				numTimeSlots++;
				elem = new Element("TIME_SLOT");
				elem.setAttribute("TIME_SLOT_ID", "ts" + numTimeSlots);
				elem.setAttribute("TIME_VALUE", "" + annotation.getEndTime());
				timeOrder.addContent(elem);
				numTimeSlots++;
			}
		}
		root.addContent(timeOrder);
		
		// build tiers
		numAnnotations = 1;
		for (ElanTier tier : this.tiers) {
			Element tierElem = new Element("TIER");
			if (tier.getID() != null) tierElem.setAttribute("TIER_ID", tier.getID());
			if (tier.getParticipant() != null) tierElem.setAttribute("PARTICIPANT", tier.getParticipant());
			if (tier.getAnnotator() != null) tierElem.setAttribute("ANNOTATOR", tier.getAnnotator());
			tierElem.setAttribute("LINGUISTIC_TYPE_REF", "default-lt");
			tierElem.setAttribute("DEFAULT_LOCALE", "en");
			List<ElanAnnotation> annotations = tier.getAnnotations();
			for (ElanAnnotation annotation : annotations) {
				
				Element elem = new Element("ANNOTATION_VALUE");
				elem.setText(annotation.getValue());
				elem = new Element("ALIGNABLE_ANNOTATION").addContent(elem);
				elem.setAttribute("ANNOTATION_ID", "a" + numAnnotations);
				elem.setAttribute("TIME_SLOT_REF1", "ts" + ((numAnnotations*2)-1));
				elem.setAttribute("TIME_SLOT_REF2", "ts" + (numAnnotations*2));
				elem = new Element("ANNOTATION").addContent(elem);
				tierElem.addContent(elem);
				numAnnotations++;
			}
			root.addContent(tierElem);
		}
		
		// build footer
		root.addContent(new Element("LINGUISTIC_TYPE").setAttribute("LINGUISTIC_TYPE_ID", "default-lt").setAttribute("TIME_ALIGNABLE", "true").setAttribute("GRAPHIC_REFERENCES", "false"));
		root.addContent(new Element("LOCALE").setAttribute("LANGUAGE_CODE", "en").setAttribute("COUNTRY_CODE", "US"));
		root.addContent(new Element("CONSTRAINT").setAttribute("STEREOTYPE", "Time_Subdivision").setAttribute("DESCRIPTION", "Time subdivision of parent annotation's time interval, no time gaps allowed within this interval"));
		root.addContent(new Element("CONSTRAINT").setAttribute("STEREOTYPE", "Symbolic_Subdivision").setAttribute("DESCRIPTION", "Symbolic subdivision of a parent annotation. Annotations refering to the same parent are ordered"));
		root.addContent(new Element("CONSTRAINT").setAttribute("STEREOTYPE", "Symbolic_Association").setAttribute("DESCRIPTION", "1-1 association with a parent annotation"));
		root.addContent(new Element("CONSTRAINT").setAttribute("STEREOTYPE", "Included_In").setAttribute("DESCRIPTION", "Time alignable annotations within the parent annotation's time interval, gaps are allowed"));
		
		
		// write document to output stream
		XMLOutputter outputter = new XMLOutputter();
        outputter.setFormat(Format.getPrettyFormat());
        outputter.output(doc, outputStream);
	}
	
	/**
	 * @return the mediaFiles
	 */
	public List<ElanLinkedFile> getMediaFiles() {
		return mediaFiles;
	}

	/**
	 * @param index
	 * @return the mediaFile at index.
	 */
	public ElanLinkedFile getMediaFile(int index) {
		return mediaFiles.get(index);
	}
	
	/**
	 * @param mediaFiles the mediaFiles to set
	 */
	public void setMediaFiles(List<ElanLinkedFile> mediaFiles) {
		this.mediaFiles = mediaFiles;
	}
	
	/**
	 * @param index
	 * @param mediaFile the mediaFile to set
	 */
	public void setMediaFile(int index, ElanLinkedFile mediaFile) {
		this.mediaFiles.set(index, mediaFile);
	}
	
	/**
	 * @param mediaFile the mediaFile to add
	 */
	public void addMediaFile(ElanLinkedFile mediaFile) {
		this.mediaFiles.add(mediaFile);
	}
	
	/**
	 * @param index
	 */
	public void removeMediaFile(int index) {
		this.mediaFiles.remove(index);
	}

	/**
	 * @return the linkedFiles
	 */
	public List<ElanLinkedFile> getLinkedFiles() {
		return linkedFiles;
	}

	/**
	 * @param index
	 * @return the linkedFile at index.
	 */
	public ElanLinkedFile getLinkedFile(int index) {
		return linkedFiles.get(index);
	}
	
	/**
	 * @param linkedFiles the linkedFiles to set
	 */
	public void setLinkedFiles(List<ElanLinkedFile> linkedFiles) {
		this.linkedFiles = linkedFiles;
	}
	
	/**
	 * @param index
	 * @param linkedFile the linkedFile to set
	 */
	public void setLinkedFile(int index, ElanLinkedFile linkedFile) {
		this.linkedFiles.set(index, linkedFile);
	}
	
	/**
	 * @param linkedFile the linkedFile to add
	 */
	public void addLinkedFile(ElanLinkedFile linkedFile) {
		this.linkedFiles.add(linkedFile);
	}
	
	/**
	 * @param index
	 */
	public void removeLinkedFile(int index) {
		this.linkedFiles.remove(index);
	}
	
	/**
	 * @return the tiers
	 */
	public List<ElanTier> getTiers() {
		return tiers;
	}

	/**
	 * @param index
	 * @return the tier at index.
	 */
	public ElanTier getTier(int index) {
		return tiers.get(index);
	}
	

	/**
	 * @param id
	 * @return the first tier with given ID.
	 */
	public ElanTier getTierByID(String id) {
		for (ElanTier tier : this.tiers)
			if (tier.getID().equals(id)) return tier;
		return null;
	}
	
	/**
	 * @param participant
	 * @return the first tier with given participant.
	 */
	public ElanTier getTierByParticipant(String participant) {
		for (ElanTier tier : this.tiers)
			if (tier.getParticipant().equals(participant)) return tier;
		return null;
	}
	
	/**
	 * @param annotator
	 * @return the first tier with given annotator.
	 */
	public ElanTier getTierByAnnotator(String annotator) {
		for (ElanTier tier : this.tiers)
			if (tier.getAnnotator().equals(annotator)) return tier;
		return null;
	}
	
	/**
	 * @param tiers the tiers to set
	 */
	public void setTiers(List<ElanTier> tiers) {
		this.tiers = tiers;
	}
	
	/**
	 * @param index
	 * @param tier the tier to set
	 */
	public void setTier(int index, ElanTier tier) {
		this.tiers.set(index, tier);
	}
	
	/**
	 * @param tier the tier to add
	 */
	public void addTier(ElanTier tier) {
		this.tiers.add(tier);
	}
	
	/**
	 * @param index
	 */
	public void removeTier(int index) {
		this.tiers.remove(index);
	}

	/**
	 * @param id
	 * @return the first tier with given ID.
	 */
	public void removeTierByID(String id) {
		for (int i = 0; i < this.tiers.size(); i++)
			if (this.tiers.get(i).getID().equals(id)) this.tiers.remove(i);
	}
	

	public static class ElanLinkedFile
	{
		private String URL = "";
		private String relativeURL = "";
		private String mimeType = "";
		private long timeOrigin = 0l;
		
		/**
		 * Constructs a new LinkedFile object.
		 */
		public ElanLinkedFile() {
			super();
		}
		
		/**
		 * Constructs a new LinkedFile object.
		 * 
		 * @param URL
		 * @param relativeURL
		 * @param mimeType
		 * @param timeOrigin
		 */
		public ElanLinkedFile(String uRL, String relativeURL, String mimeType, long timeOrigin) {
			super();
			URL = uRL;
			this.relativeURL = relativeURL;
			this.mimeType = mimeType;
			this.timeOrigin = timeOrigin;
		}

		/**
		 * @return the URL
		 */
		public String getURL() {
			return URL;
		}
		/**
		 * @param URL the URL to set
		 */
		public void setURL(String uRL) {
			URL = uRL;
		}
		/**
		 * @return the relativeURL
		 */
		public String getRelativeURL() {
			return relativeURL;
		}
		/**
		 * @param relativeURL the relativeURL to set
		 */
		public void setRelativeURL(String relativeURL) {
			this.relativeURL = relativeURL;
		}
		/**
		 * @return the mimeType
		 */
		public String getMimeType() {
			return mimeType;
		}
		/**
		 * @param mimeType the mimeType to set
		 */
		public void setMimeType(String mimeType) {
			this.mimeType = mimeType;
		}
		/**
		 * @return the timeOrigin
		 */
		public long getTimeOrigin() {
			return timeOrigin;
		}
		/**
		 * @param timeOrigin the timeOrigin to set
		 */
		public void setTimeOrigin(long timeOrigin) {
			this.timeOrigin = timeOrigin;
		}
		
	}

	public static class ElanAnnotation
	{
		private String value = "";
		private long startTime = 0l;
		private long endTime = 0l;
		
		/**
		 * Constructs a new ElanAnnotation object.
		 */
		public ElanAnnotation() {
			super();
		}
		
		/**
		 * Constructs a new ElanAnnotation object.
		 * 
		 * @param value
		 * @param startTime
		 * @param endTime
		 */
		public ElanAnnotation(String value, long startTime, long endTime) {
			super();
			this.value = value;
			this.startTime = startTime;
			this.endTime = endTime;
		}
		
		/**
		 * @return the value
		 */
		public String getValue() {
			return value;
		}
		/**
		 * @param value the value to set
		 */
		public void setValue(String value) {
			this.value = value;
		}
		/**
		 * @return the startTime
		 */
		public long getStartTime() {
			return startTime;
		}
		/**
		 * @param startTime the startTime to set
		 */
		public void setStartTime(long startTime) {
			this.startTime = startTime;
		}
		/**
		 * @return the endTime
		 */
		public long getEndTime() {
			return endTime;
		}
		/**
		 * @param endTime the endTime to set
		 */
		public void setEndTime(long endTime) {
			this.endTime = endTime;
		}
	}

	public static class ElanTier
	{
		private String ID = "";
		private String participant = "";
		private String annotator = "";
		
		private List<ElanAnnotation> annotations;

		/**
		 * Constructs a new ElanTier object.
		 */
		public ElanTier() {
			super();
			
			this.annotations = new ArrayList<ElanAnnotation>();
		}
		
		/**
		 * Constructs a new ElanTier object.
		 * 
		 * @param iD
		 */
		public ElanTier(String iD) {
			super();
			ID = iD;
			
			this.annotations = new ArrayList<ElanAnnotation>();
		}
		
		/**
		 * Constructs a new ElanTier object.
		 * 
		 * @param iD
		 * @param participant
		 * @param annotator
		 */
		public ElanTier(String iD, String participant, String annotator) {
			super();
			ID = iD;
			this.participant = participant;
			this.annotator = annotator;
			
			this.annotations = new ArrayList<ElanAnnotation>();
		}
		
		/**
		 * @return the annotations
		 */
		public List<ElanAnnotation> getAnnotations() {
			return annotations;
		}

		/**
		 * @param index
		 * @return the annotation at index.
		 */
		public ElanAnnotation getAnnotation(int index) {
			return annotations.get(index);
		}
		
		/**
		 * @return the annotation names without duplicates
		 */
		public String[] getUniqAnnotations() {
			ArrayList<String> templ = new ArrayList<String>();
			
			for (ElanAnnotation annotation : this.annotations)
				if (!templ.contains(annotation.value)) templ.add(annotation.value);
					
			String[] tempa = new String[templ.size()];
			for (int i = 0; i < templ.size(); i++) tempa[i] = templ.get(i);  
			
			return tempa;
		}
		

		/**
		 * @param value
		 * @return the first annotation with given value.
		 */
		public ElanAnnotation getAnnotationByValue(String value) {
			for (ElanAnnotation annotation : this.annotations)
				if (annotation.value.equals(value)) return annotation;
			return null;
		}
		
		/**
		 * @param timestamp
		 * @return the annotation at the given time stamp.
		 */
		public ElanAnnotation getAnnotationByTimeStamp(long timestamp) {
			if (timestamp < 0) return null;
			for (ElanAnnotation annotation : this.annotations)
				if ((annotation.startTime <= timestamp) && (annotation.endTime >= timestamp)) return annotation;
			return null;
		}
		
		/**
		 * @param annotations the annotations to set
		 */
		public void setAnnotations(List<ElanAnnotation> annotations) {
			this.annotations = annotations;
		}
		
		/**
		 * @param index
		 * @param annotation the annotation to set
		 */
		public void setAnnotation(int index, ElanAnnotation annotation) {
			this.annotations.set(index, annotation);
		}
		
		/**
		 * @param annotation the annotation to add
		 */
		public void addAnnotation(ElanAnnotation annotation) {
			this.annotations.add(annotation);
		}
		
		/**
		 * @param index
		 */
		public void removeAnnotation(int index) {
			this.annotations.remove(index);
		}

		/**
		 * @return the iD
		 */
		public String getID() {
			return ID;
		}

		/**
		 * @param iD the iD to set
		 */
		public void setID(String iD) {
			ID = iD;
		}

		/**
		 * @return the participant
		 */
		public String getParticipant() {
			return participant;
		}

		/**
		 * @param participant the participant to set
		 */
		public void setParticipant(String participant) {
			this.participant = participant;
		}

		/**
		 * @return the annotator
		 */
		public String getAnnotator() {
			return annotator;
		}

		/**
		 * @param annotator the annotator to set
		 */
		public void setAnnotator(String annotator) {
			this.annotator = annotator;
		}
	}
}
