source("defines.R")
source("EvaPar.R")
require(stringr)
performance <- function(cfm){
  ratios <- matrix(0,nrow=dim(cfm)[1], ncol=11, dimnames=list(rownames(cfm),c("tp","fp","tn","fn","n","accuracy","precision","recall","specificity","dhr","f1")))
  
  for (i in c(1:dim(cfm)[1])) {
    ratios[i,"tp"] <- tp <- cfm[i,i]
    ratios[i,"fp"] <- fp <- sum(cfm[-i,i])
    ratios[i,"tn"] <- tn <- sum(cfm[-i,-i])
    ratios[i,"fn"] <- fn <- sum(cfm[i,-i])
    ratios[i,"n"] <- n <- sum(cfm[i,])
    
    ratios[i,"accuracy"] <- (tp + tn) / (tp + fp + tn + fn)
    ratios[i,"precision"] <- tp / (tp + fp)
    ratios[i,"recall"] <- tp / (tp + fn)
    ratios[i,"specificity"] <- tn / (tn + fp)
    ratios[i,"dhr"] <- if (n == 0) ratios[i, "specificity"] else min(ratios[i, c("recall", "specificity")])
    if (tp + fn + fp == 0)
      ## only true negatives, set F1-score to 1.0
      ratios[i, "f1"] <- 1
    else
      ratios[i,"f1"] <- 2 * tp / (2 * tp + fn + fp)
  }
  accuracy <- sum(diag(cfm))/sum(cfm)
  precision <- ratios[,"precision"]*ratios[,"n"]
  precision <- sum(precision[is.finite(precision)], na.rm=TRUE)/sum(ratios[,"n"])
  recall <- ratios[,"recall"]*ratios[,"n"]
  recall <- sum(recall[is.finite(recall)], na.rm=TRUE)/sum(ratios[,"n"])
  specificity <- ratios[,"specificity"]*ratios[,"n"]
  specificity <- sum(specificity[is.finite(specificity)], na.rm=TRUE)/sum(ratios[,"n"])
  dhr <- ratios[,"dhr"]*ratios[,"n"]
  dhr <- sum(dhr[is.finite(dhr)], na.rm=TRUE)/sum(ratios[,"n"])
  f1 <- ratios[,"f1"]*ratios[,"n"]
  f1 <- sum(f1[is.finite(f1)], na.rm=TRUE)/sum(ratios[,"n"])
  
  result <- list(m=cfm, ratios=ratios, accuracy=accuracy, precision=precision, recall=recall, specificity=specificity, dhr=dhr, f1=f1)
}

analyze_hmm <- function(configfn){
  config <- read.table(configfn, header=T, stringsAsFactors=FALSE)
  res <- apply(config, MARGIN=1, function(configR){
    resfolder <- paste0(dirname(configfn),"/",configR["EXAP"],"_model",configR["MODEL"], "_ag", configR["AGENTS"], "_loo", str_trim(as.character(configR["LOO"])), "_lol", configR["LOL"], "_np", configR["NP"], "_anno", configR["ANNO"], "_obs",configR["OBS"],"_dur",configR["DUR"],"_est", configR["EST"], "/res")
    summary <- data.frame(HEU=0, REV=1, GDW=0, BIAS=0, SAL=0, SPEC=0, REC=0, SEED=1, LL=0)
    seeds <- c(1)
    tryCatch(
      afn <- paste0("../Data/",configR["AGENTS"],"_",as.integer(configR["LOL"])+1,"_anno.txt"),
      error=function(e){
        print(paste0("Problem with ", afn))
        print(configR)
      }
    )
    
    if (configR["EST"] == 1){
      perf <- merge_internal(summary, resfolder, afn, seeds)
    }else{
      perf<-mergeAn_internal(summary, resfolder, afn, seeds, as.integer(configR["AGENTS"]))
    }
    list(folder=resfolder, afn=afn, rparameter=summary,cparameter=configR, analysis=perf)
  })
  
}

analyze <- function(configfn){
  config <- read.table(configfn, header=T, stringsAsFactors=FALSE)
  res <- apply(config, MARGIN=1, function(configR){
    resfolder <- paste0(dirname(configfn),"/",configR["EXAP"],"_model",configR["MODEL"], "_ag", configR["AGENTS"], "_loo", str_trim(as.character(configR["LOO"])), "_lol", configR["LOL"], "_np", configR["NP"], "_anno", configR["ANNO"], "_obs",configR["OBS"],"_dur",configR["DUR"],"_est", configR["EST"], "/res")
    if (!file.exists(paste0(resfolder,"/res.txt"))){#create it
      system(paste0("sh 00X-search.sh ", resfolder, " > ",resfolder, "/res.txt"))
      if(!file.exists(paste0(resfolder,"/res.txt"))){
        warning(paste0("unable to create summary file in ", resfolder))
      }
    } 
    summaryfile <- paste0(resfolder, "/res.txt");
    tryCatch(
      afn <- paste0("../Data/",configR["AGENTS"],"_",as.integer(configR["LOL"])+1,"_anno.txt"),
      error=function(e){
        print(paste0("Problem with ", afn))
        print(configR)
      }
    )
    parameter <- read.table(paste0(resfolder, "/res.txt"), header=T)
    seeds <- unique(parameter$SEED)
    if (configR["EST"] == 1){
      perf <- merge(summaryfile, afn, seeds)
    }else{
      perf<-mergeAn(summaryfile, afn, seeds, as.integer(configR["AGENTS"]))
    }
    list(folder=resfolder, summaryfile=summaryfile, afn=afn, rparameter=parameter,cparameter=configR, analysis=perf)
  })
  
}


getperfAG <- function(dens, truth, AG=1){
  d <- sapply(roomnames, function(room){
    r <- paste0("AG",AG,".",room)
    res <- dens[[r]]
    if (is.null(res)){
      res <- rep(0,dim(dens)[1])
    }
    res
  })
  est <- apply(d, which.max, MARGIN=1)
  tab <- table(rfactor[truth[1:length(est)]],rfactor[est])
  
  s <- performance(tab)
  list(perf=s, truth=rfactor[truth[1:length(est)]], est=rfactor[est])
}


merge_internal <- function(summary, resfolder, afn, seeds){
  an <- read.table(afn)
  uan <- unique(an$V3)
  numagents <- length(uan)
  truth <- lapply(uan, function(name){
    a <- an[an$V3==name,]
    as.integer(factor(a$V4,levels=levels(rfactor)))
  })
  
  
  r <- summary
  
  comb <- r[r$SEED==seeds[1],]
  a <- apply(comb, MARGIN=1, function(par){
    dseeds <- lapply(seeds, function(seed){
      name <- paste0(resfolder,"/h",par[1], "_v",par[2], "_f",par[3], "_b",par[4], "_A",par[5], "_B", par[6], "_C",par[7], "_R",seed,"_res.txt")
      print(paste0("Analyzing file ", name))
      tryCatch(
        if (file.exists(name)){
          d <- read.table(name,skip=1, col.names=c("time", colnames[1:(14*numagents)]))
        }else { # just for the case that somebody forgot to use seeds ;-)
          name <- paste0(resfolder,"/h",par[1], "_v",par[2], "_f",par[3], "_b",par[4], "_A",par[5], "_B", par[6], "_C",par[7], "_S",seed,"_res.txt")
          if (file.exists(name)){
            d <- read.table(name,skip=1,col.names=c("time", colnames[1:28]))
          }
          else{
            name <- paste0(resfolder,"/h",par[1], "_v",par[2], "_f",par[3], "_b",par[4], "_A",par[5], "_B", par[6], "_C",par[7],"_res.txt")
            if (file.exists(name)){
              d <- read.table(name,skip=1,col.names=c("time", colnames[1:28]))
            }
            warning(paste0("Use of file ",name, " without different randomseeds."))
          }
        },
        error=function(w){
          print(paste("Error in file ", name))
          print(w)
        })
    })
    b <- Reduce("+", dseeds)
    res <- lapply(1:length(truth), function(truthI){
      l<-getperfAG(b, truth[[truthI]], truthI)
      n <- sapply(1:length(l$est), function(I){
        simple[l$est[I],l$truth[I]] ^ 2  
      })
      
      list(perf=l$perf, N2=sum(n))
    })
    summary <- list(par=par,result=res, dist=b, truth=truth)
  })
  invisible(sapply(1:numagents, function(AgI){
    comb[paste0("AG",AgI,".accuracy")] <<- t(sapply(a, function(al){al$result[[AgI]]$perf$accuracy}))[1,]
    comb[paste0("AG",AgI,".precision")] <<- t(sapply(a, function(al){al$result[[AgI]]$perf$precision}))[1,]
    comb[paste0("AG",AgI,".specificity")] <<- t(sapply(a, function(al){al$result[[AgI]]$perf$specificity}))[1,]
    comb[paste0("AG",AgI,".recall")] <<- t(sapply(a, function(al){al$result[[AgI]]$perf$recall}))[1,]
    comb[paste0("AG",AgI,".dhr")] <<- t(sapply(a, function(al){al$result[[AgI]]$perf$dhr}))[1,]
    comb[paste0("AG",AgI,".f1")] <<- t(sapply(a, function(al){al$result[[AgI]]$perf$f1}))[1,]
    comb[paste0("AG",AgI,".N2")] <<- t(sapply(a, function(al){al$result[[AgI]]$N2}))[1,]     
  }))
  if (numagents > 1){
    comb$N2 <-rowSums(comb[,paste0("AG",1:numagents,".N2")])
  }else{
    comb$N2 <- comb$AG1.N2
  }
  comb$MSE <-comb$N2 / (numagents*length(truth[[1]]))
  
  # append performance values to combinations 
  list(overview=comb,detailed=a)
}

merge <- function(summaryfile, afn, seeds){
  r <- read.table(summaryfile, header=T)
  resfolder <- dirname(summaryfile)
  merge_internal(r, resfolder, afn, seeds) 
}

mergeAn <- function(summaryfile, afn, seeds, numagents){
  r <- read.table(summaryfile, header=T)
  resfolder <- dirname(summaryfile)
  mergeAn_internal(r,resfolder,afn,seeds,numagents)
}

mergeAn_internal <- function(summary, resfolder, afn, seeds, numagents){
   anno <- read.anno.Room(afn=afn,numagents=numagents)

  r <- summary
  comb <- r[r$SEED==seeds[1],]
  a <- apply(comb, MARGIN=1, function(par){
    dseeds <- lapply(seeds, function(seed){
      name <- paste0(resfolder,"/h",par[1], "_v",par[2], "_f",par[3], "_b",par[4], "_A",par[5], "_B", par[6], "_C",par[7], "_R",seed,"_res.txt")
      print(paste0("Analyzing file ", name))
      tryCatch(
        if (file.exists(name)){
          d <- read.table(name,header=T)
        }else {
          warning(paste0("Missing file ",name))
        },
        error=function(w){
          print(paste("Error in file ", name))
          print(w)
        })
      if (dim(d)[1] != length(anno[[1]])){
        m <- as.data.frame(matrix(0, ncol=dim(d)[2], nrow=length(anno[[1]]) - dim(d)))
        names(m) <- names(d)
        d <- rbind(d, m)
      }
      d
    })
    b <- Reduce("+", dseeds)
    #anno was read with read.anno.ROOM
    res <- lapply(1:length(anno), function(truthI){
      ev<- get.room.dist(b, truthI, anno)
      tab <- table(ev$anno, ev$est)
      n <- sum(abs(as.integer(ev$anno) - as.integer(ev$est)))
      n2 <- sum((as.integer(ev$anno) - as.integer(ev$est))^2)
      s <- performance(tab)
      ev$num <- n
      ev$num2 <- n2
      ev$perf <- s
      ev
    })
    summary <- list(par=par, result=res, dist=b, truth=anno)
    })
   
  invisible(sapply(1:14, function(AgI){
    comb[paste0(roomnames[AgI],".accuracy")] <<- t(sapply(a, function(al){al$result[[AgI]]$perf$accuracy}))[1,]
    comb[paste0(roomnames[AgI],".precision")] <<- t(sapply(a, function(al){al$result[[AgI]]$perf$precision}))[1,]
    comb[paste0(roomnames[AgI],".specificity")] <<- t(sapply(a, function(al){al$result[[AgI]]$perf$specificity}))[1,]
    comb[paste0(roomnames[AgI],".recall")] <<- t(sapply(a, function(al){al$result[[AgI]]$perf$recall}))[1,] 
    comb[paste0(roomnames[AgI],".dhr")] <<- t(sapply(a, function(al){al$result[[AgI]]$perf$dhr}))[1,] 
    comb[paste0(roomnames[AgI],".f1")] <<- t(sapply(a, function(al){al$result[[AgI]]$perf$f1}))[1,] 
    comb[paste0(roomnames[AgI],".N")] <<- t(sapply(a, function(al){al$result[[AgI]]$num}))[1,]  
    comb[paste0(roomnames[AgI],".N2")] <<- t(sapply(a, function(al){al$result[[AgI]]$num2}))[1,]  
  }))
  comb$N2 <-rowSums(comb[,paste0(roomnames,".N2")])
  comb$MSE <-comb$N2 / (length(anno)*length(anno[[1]]))
   
  # append performance values to combinations 
  list(overview=comb,detailed=a)
}

getperf <- function(dens, truth){
  d <- sapply(roomnames, function(room){
    res <- dens[[room]]
    if (is.null(res)){
      res <- rep(0,dim(dens)[1])
    }
    res
  })
  est <- apply(d, which.max, MARGIN=1)
  tab <- table(rfactor[truth[1:length(est)]],rfactor[est])
  s <- performance(tab)
}

plotcomp <- function(comb){
  plot(comb$LL,comb$accuracy, cex=.1, ylim=(c(.3,1)))
  points(comb$LL, comb$precision, col="red", cex=.1)
  points(comb$LL, comb$specificity, col="blue", cex=.1)
  legend("bottom", c("accuracy", "precision", "specificity"), col=c("black","red","blue"), pch=c(1,1,1),pt.cex=.3)
}

createfromfile <- function(ofn, afn, show=F){
  d <- read.table(ofn,header=T)
  a <- read.table(afn)
  truth <- as.integer(factor(a$V4,levels=levels(rfactor)))
  
  s <- getperf(d,truth)
  
  
  if (show){
    op  <- par(mfrow=c(2,1), mar=c(0,3,2.5,0),cex.axis=0.5)
    
    call <- paste0(match.call(),collapse='')
    image(1- as.matrix(d), yaxt='n', xaxt='n', xlab="Time", main=paste("acc=", s$accuracy,", prec=",s$precision,", \nrecall=",s$recall, ", spec=",s$specificity,",\ncall=",call), cex.main=0.8,col=gray.colors(256))
    axis(labels=rfactor, at=seq(0,1,length.out=length(rooms)), side=2,las=2)
    lines(seq(0,1,length.out=length(est)),(est-1)/(length(rooms)-1), lty=1)
    lines(seq(0,1,length.out=length(est)), (truth-1)/(length(rooms)-1), type="l", lty=2, lwd=2)
    points(seq(0,1,length.out=length(est)),1-(truth[1:length(est)] == est)-1 , col="red",cex=.2)
    par(mar=c(0.1,0,0,0))
    legend("bottom",c("truth", "estimate"),  lty=c(2,1))
    plot(tab, las=2,main="")
    par(op)
  }
  invisible(list(estimate=est, truth=truth, perf=s))
}


createfromfile <- function(ofn, range, afn){
  d <- read.table(ofn,header=T)
  rooms <- c("corr1", "corr2", "corr3", "corr4", "corr5", "room207", "room208", "room210", "room211", "room205", "room214a", "room218", "stairs", "WC")
  rfactor <- factor(rooms, levels=rooms)
  d <- sapply(rooms, function(room){
    res <- d[[room]]
    if (is.null(res)){
      res <- rep(0,dim(d)[1])
    }
    res
  })
# we need one more column
est <- apply(d, which.max, MARGIN=1)
#plot(est)
a <- read.table(afn)
truth <- as.integer(factor(a$V4,levels=levels(rfactor)))

tab <- table(rfactor[truth[1:length(est)]],rfactor[est])
s <- performance(tab)
op  <- par(mfrow=c(2,1), mar=c(0,3,2.5,0),cex.axis=0.5)

call <- paste0(match.call(),collapse='')
image(1- as.matrix(d), yaxt='n', xaxt='n', xlab="Time", main=paste("acc=", s$accuracy,", prec=",s$precision,", \nrecall=",s$recall, ", spec=",s$specificity,",\ncall=",call), cex.main=0.8,col=gray.colors(256))
#image(1- as.matrix(d[range]))
axis(labels=rfactor, at=seq(0,1,length.out=length(rooms)), side=2,las=2)
lines(seq(0,1,length.out=length(est)),(est-1)/(length(rooms)-1), lty=1)

lines(seq(0,1,length.out=length(est)), (truth-1)/(length(rooms)-1), type="l", lty=2, lwd=2)
points(seq(0,1,length.out=length(est)),1-(truth[1:length(est)] == est)-1 , col="red",cex=.2)
par(mar=c(0.1,0,0,0))
plot(tab, las=2,main="")


#par(op)
#legend("bottom",c("truth", "estimate"),  lty=c(2,1))


invisible(list(estimate=est, truth=truth, perf=s))
}

roomDist <- function(ofn, room, ranno){
  data <- read.table(ofn)
  get.room.dist(data,room,ranno)
  
}

get.room.dist <- function(data, room,ranno){
  scol <- (1:8) + ((as.integer(room)-1) * 8)
  m <- as.matrix(data[,scol])
  est <- sapply(1:dim(m)[1], function(Irow){which.max(m[Irow,])-1})
  numf <- factor(0:8)
  list(dist=m, est=factor(est, levels=levels(numf)), anno=factor(ranno[[room]], levels=levels(numf)), room=roomnames[room])
}

readMAanno <- function(dataset = 2, iteration = 2){
  a <- paste0("../Data/",dataset,"_",iteration,"_anno.txt")
  read.anno.Room(a, dataset)
}

read.anno.Room <-  function(afn, numagents){
  name <- basename(afn)
  if(exists(name)){
      return (get(name))
  }
  #change to docall cbind
  a <- read.table(afn)
  # we can do that because we have <dataset> number of agents
  l <- dim(a)[1]/numagents
  anno <- a[1:l,]
  if (numagents > 1){
    for (i in 2:numagents){
      c <- ((l*(i-1))+1):((l*(i-1))+l)
      anno <- cbind(anno, a[c,])
    }
  }
  ranno <- lapply(roomnames, function(room){
    rowSums(
      t(
        sapply(1:dim(anno)[1], function(Irow){
          anno[Irow,]==room
        })
      )
    )
  })
  assign(name,ranno,.GlobalEnv)
  get(name)
}

plotroomdist <- function(ev){
  tab <- table(ev$anno, ev$est)
  s <- performance(tab)
  
  image(ev$dist, col=gray.colors(256), main=paste("Room=", ev$room, "\nacc=", s$accuracy,", prec=",s$precision,", \nrecall=",s$recall, ", spec=",s$specificity))
  l <- dim(ev$dist)[1]
  lines(seq(0,1,length.out=l), (as.integer(ev$est)-1)/8)
  lines(seq(0,1,length.out=l), (as.integer(ev$anno)-1)/8, lty=2)
  
}

#example for MA Evaluation
#sapply(1:14, function(I){plotroomdist(roomDist("../Models/HMM/2_agenten-2-2_dist.txt",I,readMAanno(2,2)))})

#plot 2 pictures in a column

#pdf()

#createfromfile(ofn="../HMM/out_new2.txt",range=8:17, afn="../Processed/1_5_anno.txt")
#createfromfile(ofn="../HMM/out_new.txt",range=8:17, afn="../Processed/1_5_anno.txt")
#dev.off()
#restore original layout

convert <- function(res){
  est0 <- NULL
  est1 <- list()
  
  l <- lapply(1:length(res), function(sI){
    if (res[[sI]]$cparameter["EST"]==0){
      if(is.null(est0)){
        est0 <<- base::merge(data.frame(matrix(res[[sI]]$cparameter, nrow=1)), res[[sI]]$analysis$overview)
      }else{
        est0 <<- rbind(est0, base::merge(data.frame(matrix(res[[sI]]$cparameter, nrow=1)), res[[sI]]$analysis$overview))
      }
    }
    else{
      nagent <- res[[sI]]$cparameter["AGENTS"]
      if (is.null(est1[[paste0("AG",nagent)]])){
        est1[[paste0("AG",nagent)]] <<- base::merge(data.frame(matrix(res[[sI]]$cparameter, nrow=1)), res[[sI]]$analysis$overview)
      }else{
        est1[[paste0("AG",nagent)]] <<- rbind(est1[[paste0("AG",nagent)]], base::merge(data.frame(matrix(res[[sI]]$cparameter, nrow=1)), res[[sI]]$analysis$overview))  
      }
    }
  })

  
  n <- names(est0)
  n[1:length(res[[1]]$cparameter)] <- names(res[[1]]$cparameter)
  names(est0) <- n
  
  if (0 < length(est1)){
    for (i in 1:length(est1)){
      n <- names(est1[[i]])
      n[1:length(res[[1]]$cparameter)] <- names(res[[1]]$cparameter)
      names(est1[[i]]) <- n
    }
  }
  
  list(est0,est1)
}


plotboxes <- function(formula, data,inv=FALSE,...){
  s <- boxplot(formula, data=data, plot=FALSE)
  means <- s$stats[3,]
  means[is.na(means)] = mean(means[!is.na(means)])
  num <- dim(s$stats)[2]
  min <- min(means)
  max <- max(means)
  
  n = (num - 1)/ (max - min)
  m = 1 - min*((num - 1)/(max - min))
  
  tcols <- n*means + m
  if (inv == TRUE){
    tcols <- num - tcols
  }
  cols <- heat.colors(num)[tcols]
  boxplot(formula, data=data,col=cols,...)
}
